 /*
      This file is part of GNUnet

      GNUnet is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published
      by the Free Software Foundation; either version 2, or (at your
      option) any later version.

      GNUnet is distributed in the hope that it will be useful, but
      WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
      General Public License for more details.

      You should have received a copy of the GNU General Public License
      along with GNUnet; see the file COPYING.  If not, write to the
      Free Software Foundation, Inc., 59 Temple Place - Suite 330,
      Boston, MA 02111-1307, USA.
 */
 
/**
 * @file module/dht.c
 * @brief definition of the entry points to the module; implements
 *   the client-server application using the DHT service; the DHT
 *   service is based on RPC and the DHT itself is roughly based
 *   on kademlia.
 * @author Marko Rih, Christian Grothoff
 *
 * todo:
 * - implement findNodes
 * - implement dht_get
 * - implement dht_put
 * - implement ping
 * - implement rpc-findNode
 * - implement rpc-findValue
 * - implement rpc-store
 * - implement dhtMaintainJob
 * - implement dht-remove
 * - implement rpc-remove
 * - implement cs module
 * - implement RCP-lib
 * - implement tool dht-join
 * - implement tool dht-update
 * - test!
 */

#include "platform.h"
#include "gnunet_util.h"
#include "gnunet_core.h"
#include "gnunet_rpc_service.h"
#include "gnunet_dht_service.h"
#include "gnunet_dht_datastore_memory.h"

#define DEBUG_DHT YES

/**
 * Global core API.
 */
static CoreAPIForApplication * coreAPI = NULL;

/**
 * RPC API
 */
static RPC_ServiceAPI * rpcAPI = NULL;

/**
 * Number of replications for the master table.  Rather high since
 * that table is quite important.
 */
#define ALPHA 8

/**
 * Frequency of the DHT maintain job (trade-off between
 * more smooth traffic from the maintain job and useless
 * CPU consumption for the job going over the table doing
 * nothing).
 */
#define DHT_MAINTAIN_FREQUENCY (15 * cronSECONDS)

/* ********************* DHT routing ************************** */

/**
 * Per-peer information.
 */
typedef struct {
  /**
   * What was the last time we received a message from this peer?
   */
  cron_t lastActivity;
  /**
   * What was the last time we received a table status message
   * from this peer?
   */
  cron_t lastTableRefresh;
  /**
   * In which tables do we know that peer to participate in?
   */
  DHT_TableId * tables;
  /**
   * How large is the tables array?
   */
  unsigned int tableCount;
  /**
   * What is the identity of the peer?
   */
  HostIdentity id;
} PeerInfo;

/**
 * Peers are grouped into buckets.
 */
typedef struct {
  /**
   * Peers in this bucket fall into the distance-range 
   * (2^bstart to 2^bend].
   */
  unsigned int bstart;

  /**
   * Peers in this bucket fall into the distance-range 
   * (2^bstart to 2^bend].
   */
  unsigned int bend;

  /**
   * Peers in this bucket.  NULL is used if no peer is known.
   */
  Vector * peers; /* contains PeerInfo instances */
} PeerBucket;

/**
 * The buckets (Kademlia style routing table).
 */
static PeerBucket * buckets;

/**
 * Total number of active buckets.
 */
static unsigned int bucketCount;   

/**
 * The ID of the master table.
 */
static HashCode160 masterTableId;

/**
 * Local information about a DHT table that this peer is participating
 * in.
 */
typedef struct {
  DHT_TableId id;
  DHT_Datastore * store;
  int flags;
  /**
   * What was the last time we advertised this nodes participation in
   * this table to the master table?
   */ 
  cron_t lastMasterAdvertisement;
} LocalTableData;

/**
 * List of the tables that this peer participates in.
 */
static LocalTableData * tables;

/**
 * Number of entries in the tables array.
 */
static unsigned int tablesCount;

/**
 * Mutex to synchronize access to tables.
 */
static Mutex lock;

/* ********************* main DHT operations ****************** */

/**
 * Callback for findNodes that is invoked whenever a node is found.
 *
 * @param identity the identity of the node that was found
 * @return OK to continue searching, SYSERR to abort early
 */
typedef int (*NodeFoundCallback)(HostIdentity * identity,
				 void * closure);

/**
 * In the induced sub-structure for the given 'table', find the (k)
 * nodes closest to the given key.  The code first tries to find ALPHA
 * nodes in the routing table that participate in the given table.  If
 * nodes are found, the ALPHA nodes closest to the key are queried
 * (using the find node RPC) to find nodes close to the key.
 *
 * If no participating nodes are found, the a set of introduction
 * nodes for this table is obtained from the master table (using RPC
 * get).  For the master table we try to discover peers participating
 * in the DHT using broadcasts to all connected peers (relying on
 * GNUnet core peer discovery).
 *
 * If we learn about new nodes in this step, add them to the RT table;
 * if we run out of space in the RT, send pings to oldest entry; if 
 * oldest entry did not respond to PING, replace it!
 * 
 * @param table the table which the peers must participate in
 * @param key the target key to use for routing
 * @param callback function to call for each peer found
 * @param closure extra argument to the callback
 * @param timeout how long to wait (at most, callback may also trigger termination)
 * @return number of peers found, SYSERR on error
 */
static int findNodes(const DHT_TableId * table,
		     const HashCode160 * key,
		     NodeFoundCallback callback,
		     void * closure,
		     cron_t timeout) {

  return SYSERR;
}


/**
 * Cron-job to maintain DHT invariants.  The responsibility of
 * this job is to maintain the routing table (by finding peers
 * if necessary).
 */ 
static void dhtMaintainJob(void * unused) {
  int i;

  MUTEX_LOCK(&lock);
  for (i=0;i<tablesCount;i++) {
    if (! equalsHashCode160(&tables[i].id, &masterTableId)) {
      /*a) do rpc-put on master table: add this node
        to list of nodes that participate in this table */
    }
  }
  /*
    for each table that we have joined:
     b) gather neighbours using
        findNodes(table, coreAPI->myIdentity, timeout);
  */
  /* 
     for all peers in RT:
     a) if lastTableRefresh is very old, send ping
     b) if lastActivity is very very old, drop
  */
  /* 
     for all content in all tables:
       check if this peer should still be responsible for it, if not,
       migrate!
  */
  MUTEX_UNLOCK(&lock);
} 


/**
 * Perform a synchronous GET operation on the DHT identified by
 * 'table' using 'key' as the key; store the result in 'result'.  If
 * result->dataLength == 0 the result size is unlimited and
 * result->data needs to be allocated; otherwise result->data refers
 * to dataLength bytes and the result is to be stored at that
 * location; dataLength is to be set to the actual size of the
 * result.
 *
 * The peer does not have to be part of the table!
 *
 * @param table table to use for the lookup
 * @param key the key to look up  
 * @param timeout how long to wait until this operation should
 *        automatically time-out
 * @param maxResults maximum number of results to obtain, size of the results array
 * @param results where to store the results (on success)
 * @return number of results on success, SYSERR on error (i.e. timeout)
 */
static int dht_get(const DHT_TableId * table,
		   const HashCode160 * key,
		   cron_t timeout,
		   unsigned int maxResults,
		   DHT_DataContainer ** results) {
  /* check if we are part of this table */
  int i;
  int idx;
  int ret;
  HostIdentity * peers;
  unsigned int k;

  idx = -1;
  MUTEX_LOCK(&lock);  
  for (i=0;i<tablesCount;i++) {
    if (equalsDHT_TableId(table, &tables[i].id)) {
      idx = i;
      break;
    }
  }
  /* for get/remove: if this peer participates, it cannot
     hurt to also check the local table */
  if (idx != -1) {
    /* tables[i].store->get(); */
    /* only for get: if maxResults found, no need to go further! */
  }

  /* find peers responsible for this key in table
     (using findNodes); route request there! */
  
  k = ALPHA;
  peers = MALLOC(sizeof(HostIdentity) * k);
  /* FIXME: it would be best to run findNodes and
     findValue in parallel (while findNodes still
     gathers additional nodes, send the request 
     already to the nodes discovered so far!) */
  ret = 42; /*findNodes(table,
		  key, 
		  peers,
		  k,
		  timeout);*/
  /* route get request to peers */
  /* in parallel: */
  for (i=0;i<ret;i++) {
    /* rpc_DHT_findValue(peers[i], key, maxResults, results,
       timeout); */
  }
  /* for put: */
  if (idx != -1) {
    /* if THIS peer is in the ALPHA closest peers obtained
       by findNodes, then also do local insertion! */
    /* tables[idx].store->put(); */
  }
  /* for get: if caching is on, cache values (extra data structures!) ! */
  FREE(peers);
  if (ret >= 1)
    ret = OK;
  else 
    ret = SYSERR;
  MUTEX_UNLOCK(&lock);
  return ret;
}

/**
 * Perform a synchronous put operation.   The peer does not have
 * to be part of the table!
 *
 * @param table table to use for the lookup
 * @param key the key to store
 * @param timeout how long to wait until this operation should
 *        automatically time-out
 * @param value what to store
 * @param flags bitmask
 * @return OK on success, SYSERR on error (or timeout)
 */
static int dht_put(const DHT_TableId * table,
		   const HashCode160 * key,
		   cron_t timeout,
		   const DHT_DataContainer * value,
		   int flags) {
  /* analogous to dht_get */
  return SYSERR;
}

/**
 * Perform a synchronous remove operation.  The peer does not have
 * to be part of the table!
 *
 * @param table table to use for the lookup
 * @param key the key to store
 * @param timeout how long to wait until this operation should
 *        automatically time-out
 * @param value what to remove; NULL for all values matching the key
 * @param flags bitmask
 * @return OK on success, SYSERR on error (or timeout)
 */
static int dht_remove(const DHT_TableId * table,
		      const HashCode160 * key,
		      cron_t timeout,
		      const DHT_DataContainer * value,
		      int flags) {
  /* analogous to dht_get */
  return SYSERR;
}


/**
 * Join a table (start storing data for the table).  Join
 * fails if the node is already joint with the particular
 * table.
 *
 * @param datastore the storage callbacks to use for the table
 * @param table the ID of the table
 * @param timeout how long to wait for other peers to respond to 
 *   the join request (has no impact on success or failure)
 * @param flags 
 * @return SYSERR on error, OK on success
 */
static int dht_join(DHT_Datastore * datastore,
		    const DHT_TableId * table,
		    cron_t timeout,
		    int flags) {
  int i;

  MUTEX_LOCK(&lock);
  for (i=0;i<tablesCount;i++) {
    if (equalsDHT_TableId(&tables[i].id, table)) {
      MUTEX_UNLOCK(&lock);
      return SYSERR;
    }
  }
  GROW(tables,
       tablesCount,
       tablesCount+1);
  tables[tablesCount-1].id = *table;
  tables[tablesCount-1].store = datastore;
  tables[tablesCount-1].flags = flags;
  MUTEX_UNLOCK(&lock);
  return OK;
}

typedef struct {
  DHT_TableId table;
  cron_t timeout;  
} MigrationClosure;

/**
 * Callback function to migrate content to other peers.
 */
static int dht_migrate(const HashCode160 * key,
		       const DHT_DataContainer * value,
		       int flags,
		       MigrationClosure * cls) {
  cron_t now;
  cronTime(&now);
  if (now >= cls->timeout) {
    LOG(LOG_DEBUG,
	"Aborting DHT migration due to timeout.\n");
    return SYSERR; /* abort: timeout */
  } 
  if (OK != dht_put(&cls->table,
		    key,
		    cls->timeout - now,
		    value,
		    flags))
    LOG(LOG_DEBUG,
	"Failed to migrate DHT content.\n");
  return OK;  
}

  
/**
 * Leave a table (stop storing data for the table).  Leave
 * fails if the node is not joint with the table.
 *
 * @param datastore the storage callbacks to use for the table
 * @param table the ID of the table
 * @param timeout how long to wait for other peers to respond to 
 *   the leave request (has no impact on success or failure);
 *   but only timeout time is available for migrating data, so
 *   pick this value with caution.
 * @param flags 
 * @return SYSERR on error, OK on success
 */
static int dht_leave(const DHT_TableId * table,
		     cron_t timeout,
		     int flags) {
  int i;
  int idx;
  LocalTableData old;
  MigrationClosure cls;

  MUTEX_LOCK(&lock);
  idx = -1;
  for (i=0;i<tablesCount;i++) {
    if (equalsDHT_TableId(&tables[i].id, table)) {
      idx = i;
      break;
    }
  }
  if (idx == -1) {
    MUTEX_UNLOCK(&lock);
    return SYSERR;
  }
  old = tables[i];
  tables[i] = tables[tablesCount-1];
  GROW(tables,
       tablesCount,
       tablesCount-1);
  MUTEX_UNLOCK(&lock);

  /* migrate content if applicable! */
  if ((flags & DHT_FLAGS_TABLE_MIGRATION_FLAG) > 0) {
    cls.table = *table;
    cls.timeout = cronTime(NULL) + timeout;
    old.store->iterate(old.store->closure,
		       0,
		       (DHT_DataProcessor) &dht_migrate,
		       &cls);
  }
  if (! equalsHashCode160(&masterTableId,
			  table)) {
    /* also issue dht_remove to remove this peer
       from the master node! */
  }
  return OK;
}

/**
 * We received a PING from another DHT.  The appropriate response
 * is to send a list of the tables that this peer participates in.
 */
static void rpc_DHT_ping(RPC_Param * arguments,
			 RPC_Param * results) {
}

/**
 * Find nodes that we know of that participate in the given
 * table and that are close to the given key.
 * @param arguments::key the key to route towards
 * @param arguments::timeout how long to wait at most for a reply
 * @param arguments::table the id of the table
 * @param results::peers list of peers found to participate in the given table with ID close to key
 * @param tables::list of tables that this peer participates in (optional)
 */
static void rpc_DHT_findNode(RPC_Param * arguments,
			     RPC_Param * results) {
  /* maps to findNodes */
}

static void rpc_DHT_findValue(RPC_Param * arguments,
			      RPC_Param * results) {
  /* maps to dht_get  */
}

static void rpc_DHT_store(RPC_Param * arguments,
			  RPC_Param * results) {
  /* maps to dht_put  */
}

static void rpc_DHT_remove(RPC_Param * arguments,
			   RPC_Param * results) {
  /* maps to dht_remove */
}


/**
 * Handle for the masterTable datastore that is used by this node
 * to store information about which peers participate in which
 * tables (the masterTable is another DHT, this store is just the
 * part of the masterTable that is stored at this peer).
 */
static DHT_Datastore * masterTableDatastore;

/**
 * Provide the DHT service.  The DHT service depends on the RPC
 * service.
 *
 * @param capi the core API
 * @return NULL on errors, DHT_API otherwise
 */
DHT_ServiceAPI * provide_dht_protocol(CoreAPIForApplication * capi) {
  static DHT_ServiceAPI api;
  unsigned int i;
  

  coreAPI = capi;
  rpcAPI = capi->requestService("rpc");
  if (rpcAPI == NULL)
    return NULL;
  i = getConfigurationInt("DHT",
			  "BUCKETCOUNT");
  if (i == 0)
    i = 160;
  GROW(buckets,
       bucketCount,
       i);
  for (i=0;i<bucketCount;i++) {
    buckets[i].bstart = 160 * i / bucketCount + 1;
    buckets[i].bend = 160 * (i+1) / bucketCount;
    buckets[i].peers = vectorNew(4);
  }
  
  rpcAPI->RPC_register("DHT_ping",
		       &rpc_DHT_ping);
  rpcAPI->RPC_register("DHT_findNode",
		       &rpc_DHT_findNode);
  rpcAPI->RPC_register("DHT_findValue",
		       &rpc_DHT_findValue);
  rpcAPI->RPC_register("DHT_store",
		       &rpc_DHT_store);
  rpcAPI->RPC_register("DHT_remove",
		       &rpc_DHT_remove);
  MUTEX_CREATE_RECURSIVE(&lock);
  api.get = &dht_get;
  api.put = &dht_put;
  api.remove = &dht_remove;
  api.join = &dht_join;
  api.leave = &dht_leave;
  memset(&masterTableId, 0, sizeof(HashCode160));
  /* join the master table */
  i = getConfigurationInt("DHT",
			  "MASTER-TABLE-SIZE");
  if (i == 0)
    i = 65536; /* 64k memory should suffice */
  masterTableDatastore 
    = create_datastore_memory(i);
  dht_join(masterTableDatastore,
	   &masterTableId,
	   0,
	   0);
  addCronJob(&dhtMaintainJob,
	     0,
	     DHT_MAINTAIN_FREQUENCY,
	     NULL);
  return &api;
}

int release_dht_protocol() {
  unsigned int i;
  PeerInfo * bucket;

  delCronJob(&dhtMaintainJob,
	     DHT_MAINTAIN_FREQUENCY,
	     NULL);
  /* leave the master table */
  dht_leave(&masterTableId,
	    0,
	    0);
  destroy_datastore_memory(masterTableDatastore);
  rpcAPI->RPC_unregister("DHT_ping",
			 &rpc_DHT_ping);
  rpcAPI->RPC_unregister("DHT_store",
			 &rpc_DHT_store);
  rpcAPI->RPC_unregister("DHT_findNode",
			 &rpc_DHT_findNode);
  rpcAPI->RPC_unregister("DHT_findValue",
			 &rpc_DHT_findValue);
  coreAPI->releaseService(rpcAPI);
  MUTEX_DESTROY(&lock);
  rpcAPI = NULL;
  coreAPI = NULL;

  for (i=0;i<bucketCount;i++) {
    bucket = (PeerInfo*) vectorGetFirst(buckets[i].peers);
    while (bucket != NULL) {
      GROW(bucket->tables,
	   bucket->tableCount,
	   0); 
      bucket = (PeerInfo*) vectorGetNext(buckets[i].peers);
    }
    vectorFree(buckets[i].peers);
  }
  GROW(buckets,
       bucketCount,
       0);

  return OK;
}


/* end of dht.c */
